import numpy as np
from numpy.fft import fft2, ifft2, fftfreq
from enum import Enum
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike
from typing import Union, Optional
import time

from ..core import Detector


class WFSResult:
    """
    Data structure for holding wavefront shaping results and statistics.

    Attributes:
        t (ndarray): Measured transmission matrix. If multiple targets were used, the first dimension(s) of `t`
            denote the columns of the transmission matrix (`a` indices), and the last dimensions(s) denote the targets,
            i.e., the rows of the transmission matrix (`b` indices).
        axis (int): Number of dimensions used for denoting a single column of the transmission matrix
            (e.g., 2 dimensions representing the x and y coordinates of the SLM pixels).
        noise_factor (ndarray): The estimated loss in fidelity caused by the limited SNR (for each target).
        amplitude_factor (ndarray): Estimated reduction of the fidelity due to phase-only modulation (for each target)
            (≈ π/4 for fully developed speckle).
        non_linearity (ndarray): Estimated deviation from a sinusoid response (TODO: experimental, untested).
        n (int): Total number of segments used in the optimization. When missing, this value is set to the number of
            elements in the first `axis` dimensions of `t`.
        estimated_optimized_intensity (ndarray): When missing, estimated intensity in the target(s) after displaying the
            wavefront correction on a perfect phase-only SLM.
        intensity_offset (Optional[ndarray]): Offset in the signal strength, as a scalar, or as one value per target.
            This is the offset that is caused by a bias in the detector signal, stray light, etc. Default value: 0.0.
    """

    def __init__(self,
                 t: np.ndarray,
                 axis: int,
                 noise_factor: ArrayLike,
                 amplitude_factor: ArrayLike,
                 non_linearity: ArrayLike,
                 n: Optional[int] = None,
                 intensity_offset: Optional[ArrayLike] = 0.0):
        """
        Args:
            t(ndarray): measured transmission matrix.
            axis(int):
                number of dimensions used for denoting a single columns of the transmission matrix
                (e.g. 2 dimensions representing the x and y coordinates of the SLM pixels)
            noise_factor(ArrayLike):
                the estimated loss in fidelity caused by the the limited snr (for each target).
            amplitude_factor(ArrayLike):
                estimated reduction of the fidelity due to phase-only modulation (for each target)
                (≈ π/4 for fully developed speckle)
            non_linearity(ArrayLike):
                estimated deviation from a sinusoid responds (TODO: experimental, untested)
            n(Optional[int]): total number of segments used in the optimization.
                when missing, this value is set to the number of elements in the first `axis` dimensions of `t`.
            intensity_offset(Optional[ArrayLike]):
                offset in the signal strength, as a scalar, or as one value per target
                this is the offset that is caused by a bias in the detector signal, stray light, etc.
                default value: 0.0

        """
        self.t = t
        self.axis = axis
        self.noise_factor = np.atleast_1d(noise_factor)
        self.n = np.prod(t.shape[0:axis]) if n is None else n
        self.amplitude_factor = np.atleast_1d(amplitude_factor)
        self.estimated_enhancement = np.atleast_1d(1.0 + (self.n - 1) * self.amplitude_factor * self.noise_factor)
        self.non_linearity = np.atleast_1d(non_linearity)
        self.intensity_offset = intensity_offset * np.ones(self.non_linearity.shape) if np.isscalar(intensity_offset) \
            else intensity_offset
        after = np.sum(np.abs(t), tuple(range(self.axis))) ** 2 * self.noise_factor + intensity_offset
        self.estimated_optimized_intensity = np.atleast_1d(after)

    def __str__(self) -> str:
        noise_warning = "OK" if self.noise_factor > 0.5 else "WARNING low signal quality."
        amplitude_warning = "OK" if self.amplitude_factor > 0.5 else "WARNING uneven contribution of optical modes."
        non_linearity_warning = "OK" if self.non_linearity < 0.3 else ("WARNING non-linear phase response, check "
                                                                       "lookup table.")
        return f"""
        Wavefront shaping results:
            noise_factor: {self.noise_factor} {noise_warning}
            amplitude_factor: {self.amplitude_factor} {amplitude_warning}
            non_linearity: {self.non_linearity} {non_linearity_warning}
            estimated_enhancement: {self.estimated_enhancement}
            estimated_optimized_intensity: {self.estimated_optimized_intensity}
            """

    def select_target(self, b) -> 'WFSResult':
        """
        Returns the wavefront shaping results for a single target

        Args:
            b(int): target to select, as integer index.
                If the target array is multi-dimensional, it is flattened before selecting the `b`-th component.

        Returns: WFSResults data for the specified target
        """
        return WFSResult(t=self.t.reshape((*self.t.shape[0:2], -1))[:, :, b],
                         axis=self.axis,
                         intensity_offset=self.intensity_offset[:][b],
                         noise_factor=self.noise_factor[:][b],
                         amplitude_factor=self.amplitude_factor[:][b],
                         non_linearity=self.non_linearity[:][b],
                         n=self.n
                         )


def analyze_phase_stepping(measurements: np.ndarray, axis: int, A: Optional[float] = None):
    """Analyzes the result of phase stepping measurements, returning matrix `t` and noise statitics

    This function assumes that all measurements were made using the same reference field `A`
    and that the phase of the modulated segment/mode is phase-stepped in equally spaced steps
    between 0 and 2π, whereas the phase of the reference field is kept constant.

    Args:
        measurements(ndarray): array of phase stepping measurements.
            The array holds measured intensities
            with the first one or more dimensions corresponding to the segments(pixels) of the SLM,
            one dimension corresponding to the phase steps,
            and the last zero or more dimensions corresponding to the individual targets
            where the feedback was measured.
        axis(int): indicates which axis holds the phase steps.
        A(Optional[float]): magnitude of the reference field.
            This value is used to correctly normalize the returned transmission matrix.
            When missing, the value of `A` is estimated from the measurements.

    With `phase_steps` phase steps, the measurements are given by

    .. math::

        I_p = \\lvert A + B \\exp(i 2\\pi p / phase_{steps})\\rvert^2,

    This function computes the Fourier transform.

    .. math::

        \\frac{1}{phase_{steps}} \\sum I_p  \\exp(-i 2\\pi p / phase_{steps}) = A^* B

    The value of A^* B for each set of measurements is stored in the `field` attribute of the return
    value.
    Other attributes hold an estimate of the signal-to-noise ratio,
    and an estimate of the maximum enhancement that can be expected
    if these measurements are used for wavefront shaping.
    """
    phase_steps = measurements.shape[axis]
    n = int(np.prod(measurements.shape[:axis]))
    # M = np.prod(measurements.shape[axis + 1:])
    segments = tuple(range(axis))

    # Fourier transform the phase stepping measurements
    t_f = np.fft.fft(measurements, axis=axis) / phase_steps

    # the offset is the 0-th Fourier component
    # the signal is the first Fourier component (lowest non-zero frequency)
    # the -1 Fourier component is just the conjugate of the first Fourier component since the signal is real.
    # all other components should be 0. If they are not, they are considered to be noise

    I_f = np.sum(np.abs(t_f) ** 2, segments)  # total intensity per Fourier component per target
    signal_energy = I_f[1, ...]
    offset_energy = I_f[0, ...]
    total_energy = np.sum(I_f, axis=0)
    if phase_steps > 3:
        noise_energy = (total_energy - 2.0 * signal_energy - offset_energy) / (phase_steps - 3)
        noise_factor = np.maximum(signal_energy - noise_energy, 0.0) / signal_energy
    else:
        noise_factor = 1.0  # cannot estimate reliably

    if phase_steps > 6 and (I_f[2, ...] + I_f[3, ...]) != 0.0:
        non_linearity = (I_f[2, ...] - I_f[3, ...]) / (I_f[2, ...] + I_f[3, ...])
    else:
        non_linearity = 0.0  # cannot estimate reliably, or I_f[2] + I_f[3] == 0 (happens in simulation)

    if A is None:  # reference field strength not known: estimate from data
        t_abs = np.abs(np.take(t_f, 1, axis=axis))
        offset = np.take(t_f, 0, axis=axis)
        a_plus_b = np.sqrt(offset + 2.0 * t_abs)
        a_minus_b = np.sqrt(offset - 2.0 * t_abs)
        A = 0.5 * np.mean(a_plus_b + a_minus_b)

    t = np.take(t_f, 1, axis=axis) / A

    # compute the effect of amplitude variations.
    # for perfectly developed speckle, and homogeneous illumination, this factor will be pi/4
    amplitude_factor = np.mean(np.abs(t), segments) ** 2 / np.mean(np.abs(t) ** 2, segments)

    return WFSResult(t, axis=axis, amplitude_factor=amplitude_factor, noise_factor=noise_factor,
                     non_linearity=non_linearity, n=n)


def signal_std(signal_with_noise: np.ndarray, noise: np.ndarray) -> float:
    """
    Compute noise corrected standard deviation of signal measurement.

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise. The noise is assumed to be uncorrelated with the
            signal, such that var(measured) = var(signal) + var(noise).
        noise:
            ND array containing only noise.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise.
    """
    return float(np.sqrt(signal_with_noise.var() - noise.var()))


def cnr(signal_with_noise: np.ndarray, noise: np.ndarray) -> float:
    """
    Compute the noise-corrected contrast-to-noise ratio of a measured signal. Contrast is computed as the standard
    deviation, corrected for noise. The noise variance is computed from a separate array, containing only noise.

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise. The noise is assumed to be uncorrelated with the
            signal, such that var(measured) = var(signal) + var(noise).
        noise:
            ND array containing only noise, e.g. a dark frame.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise.
    """
    return float(signal_std(signal_with_noise, noise) / noise.std())


def contrast_enhancement(signal_with_noise: np.ndarray, reference_with_noise: np.ndarray, noise: np.ndarray) -> float:
    """
    Compute noise corrected contrast enhancement. The noise is assumed to be uncorrelated with the signal, such that
    var(measured) = var(signal) + var(noise).

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise, e.g. image signal with shaped wavefront.
        reference_with_noise:
            ND array containing a reference signal including noise, e.g. image signal with a flat wavefront.
        noise:
            ND array containing only noise.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise.
    """
    return float(signal_std(signal_with_noise, noise) / signal_std(reference_with_noise, noise))


def cross_corr_fft2(f: np.ndarray, g: np.ndarray) -> np.ndarray:
    """
    Compute cross-correlation with a 2D Fast Fourier Transform. Note that this approach will introduce wrap-around
    artefacts.

    Args:
        f, g:
            2D arrays to be correlated.
    """
    return ifft2(fft2(f).conj() * fft2(g))


def find_pixel_shift(f: np.ndarray, g: np.ndarray) -> tuple[np.intp, ...]:
    """
    Find the pixel shift between two images by performing a 2D FFT based cross-correlation.
    """
    corr = cross_corr_fft2(f, g)  # Compute cross-correlation with fft2
    s = np.array(corr).shape  # Get shape
    index = np.unravel_index(np.argmax(corr), s)  # Find 2D indices of maximum
    pix_shift = (fftfreq(s[0], 1 / s[0])[index[0]],  # Correct negative pixel shifts
                 fftfreq(s[1], 1 / s[1])[index[1]])
    return pix_shift


def frame_correlation(f: np.ndarray, g: np.ndarray) -> float:
    """
    Compute frame correlation, i.e. inner product of two frames, normalized by product of the L2 norms,
    such that frame_correlation(f, s*f) == 1, where s is a scalar value.

    Args:
        f, g    Frames (or other arrays) to be correlated.
    """
    return np.vdot(f, g) / np.sqrt(np.vdot(f, f) * np.vdot(g, g))


class WFSController:
    """
    Controller for Wavefront Shaping (WFS) operations using a specified algorithm in the MicroManager environment.
    Manages the state of the wavefront and executes the algorithm to optimize and apply wavefront corrections, while
    exposing all these parameters to MicroManager.
    """

    class State(Enum):
        FLAT_WAVEFRONT = 0
        SHAPED_WAVEFRONT = 1

    def __init__(self, algorithm, source: Optional[Detector] = None):
        """
        Args:
            algorithm: An instance of a wavefront shaping algorithm.
            source (Detector): An image source, e.g. a camera or scanner.
        """
        self.algorithm = algorithm
        self.source = source
        self._wavefront = WFSController.State.FLAT_WAVEFRONT
        self._result = None
        self._noise_factor = None
        self._amplitude_factor = None
        self._estimated_enhancement = None
        self._non_linearity = None
        self._estimated_optimized_intensity = None
        self._snr = None  # Average SNR. Computed when wavefront is computed.
        self._optimized_wavefront = None
        self._recompute_wavefront = False
        self._feedback_enhancement = None
        self._test_wavefront = False  # Trigger to test the optimized wavefront
        self._run_troubleshooter = False  # Trigger troubleshooter
        self._frame_cnr = None
        self._contrast_enhancement = None
        self.dark_frame = None
        self.before_frame = None

    @property
    def wavefront(self) -> State:
        """
        Gets the current wavefront state.

        Returns:
            State: The current state of the wavefront, either FLAT_WAVEFRONT or SHAPED_WAVEFRONT.
        """
        return self._wavefront

    @wavefront.setter
    def wavefront(self, value):
        """
        Sets the wavefront state and applies the corresponding phases to the SLM.

        Args:
            value (State): The desired state of the wavefront to set.
        """
        self._wavefront = value
        if value == WFSController.State.FLAT_WAVEFRONT:
            self.algorithm.slm.set_phases(0.0)
        else:
            if self._recompute_wavefront or self._optimized_wavefront is None:
                # select only the wavefront and statistics for the first target
                result = self.algorithm.execute().select_target(0)
                self._optimized_wavefront = -np.angle(result.t)
                self._noise_factor = result.noise_factor
                self._amplitude_factor = result.amplitude_factor
                self._estimated_enhancement = result.estimated_enhancement
                self._non_linearity = result.non_linearity
                self._estimated_optimized_intensity = result.estimated_optimized_intensity
                self._snr = 1.0 / (1.0 / result.noise_factor - 1.0)
                self._result = result
            self.algorithm.slm.set_phases(self._optimized_wavefront)

    @property
    def noise_factor(self) -> float:
        """
        Returns:
            float: noise factor: the estimated loss in fidelity caused by the the limited snr.
        """
        return self._noise_factor

    @property
    def amplitude_factor(self) -> float:
        """
        Returns:
            float: amplitude factor: estimated reduction of the fidelity due to phase-only
            modulation (≈ π/4 for fully developed speckle)
        """
        return self._amplitude_factor

    @property
    def estimated_enhancement(self) -> float:
        """
        Returns:
            float: estimated enhancement: estimated ratio <after>/<before>  (with <> denoting
            ensemble average)
        """
        return self._estimated_enhancement

    @property
    def non_linearity(self) -> float:
        """
        Returns:
            float: non-linearity.
        """
        return self._non_linearity

    @property
    def estimated_optimized_intensity(self) -> float:
        """
        Returns:
            float: estimated optimized intensity.
        """
        return self._estimated_optimized_intensity

    @property
    def snr(self) -> float:
        """
        Gets the signal-to-noise ratio (SNR) of the optimized wavefront.

        Returns:
            float: The average SNR computed during wavefront optimization.
        """
        return self._snr

    @property
    def recompute_wavefront(self) -> bool:
        """Returns: bool that indicates whether the wavefront needs to be recomputed. """
        return self._recompute_wavefront

    @recompute_wavefront.setter
    def recompute_wavefront(self, value):
        """Sets the bool that indicates whether the wavefront needs to be recomputed. """
        self._recompute_wavefront = value

    @property
    def feedback_enhancement(self) -> float:
        """Returns: the average enhancement of the feedback, returns none if no such enhancement was measured."""
        return self._feedback_enhancement

    @property
    def frame_cnr(self) -> float:
        """Returns: the noise corrected contrast to noise ratio of the frame"""
        return self._frame_cnr

    @property
    def contrast_enhancement(self) -> float:
        """Returns: the noise corrected contrast enhancement"""
        return self._contrast_enhancement

    @property
    def test_wavefront(self) -> bool:
        """Returns: bool that indicates whether test_wavefront will be performed if set."""
        return self._test_wavefront

    @test_wavefront.setter
    def test_wavefront(self, value):
        """
        Calculates the feedback enhancement between the flat and shaped wavefronts by measuring feedback for both
        cases.

        Args:
            value (bool): True to enable test mode, False to disable.
        """
        if value:
            self.wavefront = WFSController.State.FLAT_WAVEFRONT
            feedback_flat = self.algorithm.feedback.read().copy()
            self.wavefront = WFSController.State.SHAPED_WAVEFRONT
            feedback_shaped = self.algorithm.feedback.read().copy()
            self._feedback_enhancement = float(feedback_shaped.sum() / feedback_flat.sum())

        self._test_wavefront = value

    def read_frame(self) -> np.ndarray:
        """
        Read a single frame from the feedback function.
        TODO: Distinguish WFS feedback and full frame measurement.
        TODO: why the copy()?
        """
        return self.source.read().copy()

    def set_slm_random(self):
        """
        Set a pattern of random phases to the active SLM phase patch. Useful for extinguishing the
        laser light in multi-PEF setups.
        """
        self.algorithm.slm.set_phases(2 * np.pi * np.random.rand(300, 300))

    def read_dark_frame(self) -> np.ndarray:
        """
        Read and save a dark frame.

        Note: for now, assume multi-PEF
        TODO: support 1PEF and transmission setups
        """
        self.set_slm_random()
        self.dark_frame = self.read_frame()
        self._wavefront = self.State.FLAT_WAVEFRONT
        return self.dark_frame

    def read_before_frame_flatwf(self) -> np.ndarray:
        """
        Read and save a frame before the WFS experiment.
        """
        self.wavefront = WFSController.State.FLAT_WAVEFRONT
        self.before_frame = self.read_frame()
        return self.before_frame

    def read_after_frame_flatwf(self) -> np.ndarray:
        """Read a frame after the WFS experiment with flat wavefront."""
        assert self._optimized_wavefront is not None
        self.wavefront = WFSController.State.FLAT_WAVEFRONT
        return self.read_frame()

    def read_after_frame_shapedwf(self) -> np.ndarray:
        """Read a frame after the WFS experiment with shaped wavefront."""
        assert self._optimized_wavefront is not None
        self.wavefront = WFSController.State.SHAPED_WAVEFRONT
        return self.read_frame()

    def test_setup_stability(self, wait_time_s, num_of_frames):
        """Test the setup stability by repeatedly reading frames."""
        first_frame = self.read_frame()
        pixel_shifts = np.zeros(shape=(num_of_frames, 2))
        correlations = np.zeros(shape=(num_of_frames,))

        for n in range(num_of_frames):
            # self.set_slm_random()
            time.sleep(wait_time_s)
            # self._wavefront = self.State.FLAT_WAVEFRONT

            new_frame = self.read_frame()
            pixel_shifts[n, :] = find_pixel_shift(first_frame, new_frame)
            correlations[n] = frame_correlation(first_frame, new_frame)

        return pixel_shifts, correlations

    @property
    def run_troubleshooter(self) -> bool:
        """Returns: bool that indiciates whether toubleshoot will be performed if set."""
        return self._run_troubleshooter

    @run_troubleshooter.setter
    def run_troubleshooter(self, value):
        """Runs the troubleshoot function."""
        if value:
            self.troubleshoot()
        self._run_troubleshooter = value

    def troubleshoot(self):
        """
        Run a series of basic checks to find common sources of error in a WFS experiment.
        """

        ### An outline of the to be written troubleshooting code.
        ### Each block indicates a step; i.e. a particular action or calculation, and may depend on
        ### previous steps (e.g. darkframe measurement). This indication helps with selecting the
        ### required order of steps.
        ### All steps where images are taken, must take images of the same part of the sample and
        ### some features in the image must be visible.
        ### TODO: some flags to indiciate which troubleshooting functions should run (especially ability to turn off
        ### long measurements)

        assert self.source is not None

        # Capture frames before WFS
        dark_frame = self.read_dark_frame()  # Dark frame
        before_frame = self.read_before_frame_flatwf()  # Frame before WFS

        self._frame_cnr = cnr(before_frame, dark_frame)  # Contrast to Noise Ratio

        if True:  # TODO: Add flag
            # WFS experiment
            recompute_wf_flag = self.recompute_wavefront
            self.recompute_wavefront = True
            self.wavefront = WFSController.State.SHAPED_WAVEFRONT
            self.recompute_wavefront = recompute_wf_flag

            # Capture frames after WFS
            after_frame_flatwf = self.read_after_frame_flatwf()  # After-frame flat wavefront
            after_frame_shapedwf = self.read_after_frame_shapedwf()  # After-frame shaped wavefront

            self._contrast_enhancement = contrast_enhancement(after_frame_shapedwf, after_frame_flatwf, dark_frame)

        # Test setup stability
        pixel_shifts, correlations = self.test_setup_stability(1, 5)

        ### Debugging
        import matplotlib.pyplot as plt
        plt.plot(pixel_shifts[:, 0], label='x')
        plt.plot(pixel_shifts[:, 1], label='y')
        plt.plot(correlations, label='Corr. with first')
        plt.legend()
        plt.show()

        ### === Test setup stability ===
        ### Requirement: find-image-pixel-shift function
        ### Preconfig: Flat wavefront
        ### Measurement: Snap multiple images over a long period of time
        ### Calculation: Cross-correlation between images
        ### Result: xdrift, ydrift, intensity drift, warning if >threshold
        ### Note 1: measurement should take a long time, could result in significant photobleaching
        ### Note 2: larger image size for more precise x,y cross correlation
        ### Implementation complexity: 3

        ### === Check SLM timing ===
        ### Measurement: Quick measurement, change SLM pattern, quick measurement, wait,
        ### later measurement
        ### Calculation: do quick measurement and later measurement correspond
        ### Result 1: Warning if timing seems incorrect
        ### Result 2: Plot graph
        ### Implementation complexity: 5

        ### === SLM illumination ===
        ### Requirement: Depends on it's own experiment
        ### Measurement: WFS experiment on entire SLM
        ### Calculation: amplitude from WFS (from e.g. Hadamard to SLM x,y basis)
        ### Result: SLM illumination map
        ### Implementation complexity: 4

        ### === Measure unmodulated light ===
        ### Requirement: Depends on it's own experiment
        ### Measurement: 2-mode phase stepping checkerboard.
        ### Calculation:
        ###    |A⋅exp(iθ) + B⋅exp(iφ) + C|² + b.g.noise
        ### Result: fraction of modulated and fraction of unmodulated light, warning if >threshold
        ### Note: large fraction of unmodulated light could indicate wrong illumination
        ### Implementation complexity: 7

        ### === Check calibration LUT ===
        ### Requirement: Phase stepping measurement light modulation done
        ### Calculation: Check higher phasestep frequencies. Is response cosine?
        ### Result 1: Nonlinearity value. Warning if very not cosine
        ### Result 2: Plot graph
        ### Implementation complexity: 5

        ### === Done: Quantify CNR ===
        ### Requirement: darkframe, before frame, noise corrected std function
        ### Calculation: noise corrected std / std of darkframe
        ### Result: 'measured signal when dark is noise'-SNR

        ### === Quantify SNR - with frame correlation ===
        ### Requirement: cross-corr
        ### Measurement: Snap multiple images in short time
        ### Calculation: how do consecutive frames correlate?
        ### Result: 'everything not reproducible is noise'-SNR
        ### Note: How can we distinguish from slm phase jitter?
        ### Implementation complexity: 2

        ### === Done: Quantify noise in signal - from phase stepping ===
        ### Requirement: phase-stepping measurement
        ### Calculation: from phase-stepping measurements
        ### Result: 'non-linearity in phase response is noise'-SNR

        ### === Quantify photobleaching ===
        ### Requirement: WFS experiment done, frames before and after WFS experiment
        ### Calculation: loss of intensity -> % photobleached
        ### Result: Value for amount of photobleaching
        ### Implementation complexity: 2



import numpy as np
import astropy.units as u
from astropy.units import Quantity
from scipy.ndimage import zoom
from numpy.typing import ArrayLike
import time
from typing import Sequence, Optional
from ..processors import CropProcessor
from ..core import Detector, Processor, PhaseSLM, Actuator
from ..utilities import ExtentType, get_pixel_size


class Generator(Detector):
    """A detector that returns synthetic data, simulating latency and measurement duration.

    Methods:
        uniform_noise: Static method to create a generator with uniform noise.
        gaussian_noise: Static method to create a generator with Gaussian noise.
    """

    def __init__(self, generator, data_shape: Sequence[int], pixel_size: Quantity, latency=0 * u.ms,
                 duration=0 * u.ms):
        """
        Initializes the Generator class, a subclass of Detector, that generates synthetic data.

        Args:
            generator (callable): A function that takes the data shape as input and returns the generated data.
            data_shape (tuple): The shape of the data to be generated, may have any number of dimensions.
            pixel_size (Quantity, optional): The size of each pixel in the data.
            duration (Quantity[u.ms]): Duration for which the generator simulates the data acquisition.
            latency (Quantity[u.ms]): Simulated delay before data acquisition begins.
        """
        super().__init__()
        self._latency = latency
        self._pixel_size = pixel_size
        self._data_shape = data_shape
        self._duration = duration
        self._generator = generator

    @property
    def duration(self) -> Quantity[u.ms]:
        """The duration for which the generator simulates the data acquisition."""
        return self._duration

    @duration.setter
    def duration(self, value: Quantity[u.ms]):
        self._duration = value

    @property
    def latency(self) -> Quantity[u.ms]:
        """The simulated delay before data acquisition begins."""
        return self._latency

    @latency.setter
    def latency(self, value: Quantity[u.ms]):
        self._latency = value.to(u.ms)

    @property
    def pixel_size(self) -> Quantity:
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, value):
        self._pixel_size = Quantity(value)

    @property
    def data_shape(self):
        return self._data_shape

    @data_shape.setter
    def data_shape(self, value):
        self._data_shape = value

    def _fetch(self, out: Optional[np.ndarray]) -> np.ndarray:  # noqa
        latency_s = self.latency.to_value(u.s)
        if latency_s > 0.0:
            time.sleep(latency_s)

        if out is None:
            out = self._generator(self.data_shape)
        else:
            out[...] = self._generator(self.data_shape)

        duration_s = self.duration.to_value(u.s)
        if duration_s > 0.0:
            time.sleep(duration_s)
        return out

    @staticmethod
    def uniform_noise(*args, low=0.0, high=1.0, **kwargs):
        """Static method to create a generator that produces uniform noise."""

        def generator(shape):
            return np.random.default_rng().uniform(low=low, high=high, size=shape)

        return Generator(*args, generator=generator, **kwargs)

    @staticmethod
    def gaussian_noise(*args, average=0.0, standard_deviation=1.0, **kwargs):
        """Static method to create a generator that produces Gaussian noise."""

        def generator(shape):
            return np.random.default_rng().normal(loc=average, scale=standard_deviation, size=shape)

        return Generator(*args, generator=generator, **kwargs)


class MockSource(Generator):
    """
    Detector that returns pre-set data. Also simulates latency and measurement duration.
    """

    def __init__(self, data: np.ndarray, pixel_size: Optional[Quantity] = None, extent: Optional[ExtentType] = None,
                 **kwargs):
        """
        Initializes the MockSource

        Args:
            data (np.ndarray): The pre-set data to be returned by the mock source.
            pixel_size (Quantity, optional): The size of each pixel in the data.
                If not specified, the pixel size is calculated from the extent and the data shape.
                If neither pixel size nor extent are specified, the pixel size from `data` is used, if available.
                Otherwise, pixel_size is set to None.
            extent (Optional[ExtentType]): The physical extent of the data array.
                Only used when the pixel size is not specified explicitly.
            **kwargs: Additional keyword arguments to pass to the Generator base class.
        """

        def generator(data_shape):
            assert data_shape == self._data.shape
            return self._data.copy()

        if pixel_size is None:
            if extent is not None:
                pixel_size = Quantity(extent) / data.shape
            else:
                pixel_size = get_pixel_size(data)

        if pixel_size is not None and (np.isscalar(pixel_size) or pixel_size.size == 1) and data.ndim > 1:
            pixel_size = pixel_size.repeat(data.ndim)

        super().__init__(generator=generator, data_shape=data.shape, pixel_size=pixel_size, **kwargs)
        self._data = data

    @property
    def data(self):
        """
        The pre-set data to be returned by the mock source.

        Note:
            When setting the `data` property, the `data_shape` attribute is updated accordingly.
            If the `data` property is set with an array that has the `pixel_size` metadata set,
            the `pixel_size` attribute is also updated accordingly.
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        pixel_size = get_pixel_size(value)
        if pixel_size is not None:
            self.pixel_size = pixel_size

    @property
    def data_shape(self):
        return self._data.shape


class ADCProcessor(Processor):
    """Mimics an analog-digital converter.

    At the moment, only positive input and output values are supported.
    """

    def __init__(self, source: Detector, analog_max: float = 0.0, digital_max: int = 0xFFFF,
                 shot_noise: bool = False, gaussian_noise_std: float = 0.0):
        """
        Initializes the ADCProcessor class, which mimics an analog-digital converter.

        Args:
            source (Detector): The source detector providing analog data.
            analog_max (float): The maximum analog value that can be handled by the ADC.
                If set to 0.0, each measurement will be automatically scaled so that the maximum
                value in the data set returned by `source` is converted to `digital_max`.
                Note that this means that the values from two different measurements cannot be compared quantitatively.

            digital_max (int):
                The maximum digital value that the ADC can output, default is unsigned 16-bit maximum.

            shot_noise (bool):
                Flag to determine if Poisson noise should be applied instead of rounding.
                Useful for realistically simulating detectors.

            gaussian_noise_std (float):
                If >0, add gaussian noise with std of this value to the data.
        """
        super().__init__(source)
        self._analog_max = None
        self._digital_max = None
        self._shot_noise = None
        self._gaussian_noise_std = None
        self.gaussian_noise_std = gaussian_noise_std
        self.shot_noise = shot_noise
        self.analog_max = analog_max  # check value
        self.digital_max = digital_max  # check value

    def _fetch(self, out: Optional[np.ndarray], data) -> np.ndarray:  # noqa
        """Clips the data to the range of the ADC, and digitizes the values."""
        # Add gaussian noise if requested
        if self._gaussian_noise_std > 0.0:
            dtype = data.dtype
            rng = np.random.default_rng()
            gaussian_noise = self._gaussian_noise_std * rng.standard_normal(size=data.shape)
            data = (data.astype('float64') + gaussian_noise).clip(0, 2 ** 16 - 1).astype(dtype)

        if self.analog_max == 0.0:
            max = np.max(data)
            if max > 0.0:
                data = data * (self.digital_max / np.max(data))  # auto-scale to maximum value
        else:
            data = np.clip(data * (self.digital_max / self.analog_max), 0, self.digital_max)

        if self._shot_noise:
            if out is None:
                out = np.random.poisson(data)
            else:
                out[...] = np.random.poisson(data)
        else:
            if out is None:
                out = np.rint(data).astype('uint16')
            else:
                out[...] = np.rint(data).astype('uint16')

        return out

    @property
    def analog_max(self) -> Optional[float]:
        """Maximum value that the ADC can handle as input

        This value and all higher values are converted to `digital_max`.
        When set to 0.0, the input signal is scaled automatically so that the maximum corresponds to
        `digital_max`
        """
        return self._analog_max

    @analog_max.setter
    def analog_max(self, value):
        if value < 0.0:
            raise ValueError('analog_max cannot be negative')
        self._analog_max = value

    @property
    def digital_max(self) -> int:
        """Maximum value that the ADC can output.

        Default value is 0xFFFF (16 bits)
        """
        return self._digital_max

    @digital_max.setter
    def digital_max(self, value):
        if value < 0 or value > 0xFFFF:
            raise ValueError('digital_max must be between 0 and 0xFFFF')
        self._digital_max = int(value)

    @property
    def shot_noise(self) -> bool:
        """when True, apply Poisson noise to the data instead of rounding"""
        return self._shot_noise

    @shot_noise.setter
    def shot_noise(self, value: bool):
        self._shot_noise = value

    @property
    def gaussian_noise_std(self) -> int:
        return self._gaussian_noise_std

    @gaussian_noise_std.setter
    def gaussian_noise_std(self, value: int):
        self._gaussian_noise_std = value


class MockCamera(ADCProcessor):
    """Wraps any 2-d image source as a camera.

    To implement the camera interface, in addition to the Detector interface,
    we must implement left,right,top, and bottom.
    In addition, the data should be returned as uint16.
    Conversion to uint16 is implemented in the ADCProcessor base class.
    """

    def __init__(self, source: Detector, shape: Optional[Sequence[int]] = None,
                 pos: Optional[Sequence[int]] = None, **kwargs):
        """
        Args:
            source (Detector): The source detector to be wrapped.
            shape (Optional[Sequence[int]]): The shape of the image data to be captured.
            pos (Optional[Sequence[int]]): The position on the source from where the image is captured.
            **kwargs: Additional keyword arguments to be passed to the Detector base class.

            TODO: move left-right-top-bottom to CropProcessor.
            Expose the properties of CropProcessor as properties of MockCamera automatically by copying from __dict__?
        """
        self._crop = CropProcessor(source, shape=shape, pos=pos)
        super().__init__(source=self._crop, **kwargs)

    @property
    def left(self) -> int:
        """left (int): Horizontal start position of the ROI."""
        return self._crop.pos[1]

    @left.setter
    def left(self, value: int):
        self._crop.pos = (self._crop.pos[0], value)

    @property
    def right(self) -> int:
        """right (int): Horizontal end position of the ROI."""
        return self.left + self.width

    @property
    def top(self) -> int:
        """top (int): Vertical start position of the ROI."""
        return self._crop.pos[0]

    @top.setter
    def top(self, value: int):
        self._crop.pos = (value, self._crop.pos[1])

    @property
    def bottom(self) -> int:
        """bottom (int): Vertical end position of the ROI."""
        return self.top + self.height

    @property
    def height(self) -> int:
        """Height of the ROI."""
        return self._crop.data_shape[0]

    @height.setter
    def height(self, value: int):
        self._crop.data_shape = (value, self._crop.data_shape[1])

    @property
    def width(self) -> int:
        """Width of the ROI."""
        return self._crop.data_shape[1]

    @width.setter
    def width(self, value: int):
        self._crop.data_shape = (self._crop.data_shape[0], value)

    @property
    def data_shape(self):
        return self._crop.data_shape

    @data_shape.setter
    def data_shape(self, value):
        self._crop.data_shape = value


class MockXYStage(Actuator):
    """
    Mimics an XY stage actuator
    """

    def __init__(self, step_size_x: Quantity[u.um], step_size_y: Quantity[u.um]):
        """

        Args:
            step_size_x (Quantity[u.um]): The step size in the x-direction.
            step_size_y (Quantity[u.um]): The step size in the y-direction.
        """
        super().__init__()
        self.step_size_x = step_size_x.to(u.um)
        self.step_size_y = step_size_y.to(u.um)
        self._y = 0.0 * u.um
        self._x = 0.0 * u.um

    @property
    def x(self) -> Quantity[u.um]:
        return self._x

    @x.setter
    def x(self, value: Quantity[u.um]):
        self._x = value.to(u.um)

    @property
    def y(self) -> Quantity[u.um]:
        return self._y

    @y.setter
    def y(self, value: Quantity[u.um]):
        self._y = value.to(u.um)

    def home(self):
        self._x = 0.0 * u.um
        self._y = 0.0 * u.um

    @property
    def duration(self) -> Quantity[u.ms]:
        return 0.0 * u.ms


class MockSLM(PhaseSLM, Actuator):
    """
    A mock version of a phase-only spatial light modulator.
    """

    def __init__(self, shape):
        """

        Args:
            shape (Sequence[int]): The 2D shape of the SLM.
        """
        super().__init__()
        if len(shape) != 2:
            raise ValueError("Shape of the SLM should be 2-dimensional.")

        self._back_buffer = np.zeros(shape, 'float32')
        self._monitor = MockSource(self._back_buffer, pixel_size=2.0 / np.min(shape) * u.dimensionless_unscaled)

    def update(self):
        self._start()  # wait for detectors to finish
        self._monitor.data = self._back_buffer.copy()
        self._back_buffer[:] = 0.0

    def set_phases(self, values: ArrayLike, update=True):
        # no docstring, use documentation from base class
        values = np.atleast_2d(values)
        scale = np.array(self._back_buffer.shape) / np.array(values.shape)
        # TODO: replace by cv2, with area interpolation
        zoom(values, scale, output=self._back_buffer, order=0)
        if update:
            self.update()

    @property
    def phases(self) -> np.ndarray:
        """Current phase pattern on the SLM."""
        return self._monitor.data

    def pixels(self) -> Detector:
        """Returns a `camera` that returns the current phase on the SLM.

        The camera coordinates are spanning the [-1,1] range by default."""
        return self._monitor

    @property
    def duration(self) -> Quantity[u.ms]:
        return 0.0 * u.ms



import numpy as np
from ..core import Detector, PhaseSLM
from .utilities import analyze_phase_stepping, WFSResult


class StepwiseSequential:
    """
    Class definition for stepwise sequential algorithm for wavefront shaping, as described by Vellekoop [1] [2].

    TODO: enable low-res pre-optimization (as in Vellekoop & Mosk 2007)
    TODO: modulate segments in pupil (circle) only

    [^1]:Vellekoop, I. M., & Mosk, A. P. (2007). Focusing coherent light through opaque strongly scattering media. Optics
    letters, 32(16), 2309-2311.
    [2]: Ivo M. Vellekoop, "Feedback-based wavefront shaping," Opt. Express 23, 12189-12206 (2015)
    """

    def __init__(self, feedback: Detector, slm: PhaseSLM, phase_steps=4, n_x=4, n_y=None):
        """
        This class systematically modifies the phase pattern of each SLM element and measures the resulting feedback.

        Args:
            feedback (Detector): Source of feedback
            slm (PhaseSLM): The spatial light modulator
            phase_steps (int): The number of phase steps.
            n_x (int): Number of SLM elements in x direction
            n_y (int): Number of SLM elements in y direction
        """
        self._n_x = n_x
        self._n_y = n_x if n_y is None else n_y
        self._phase_steps = phase_steps
        self.slm = slm
        self.feedback = feedback

    def execute(self) -> WFSResult:
        """
        Returns:
            WFSResult: An object containing the computed SLM transmission matrix and related data.
        """
        phase_pattern = np.zeros((self.n_y, self.n_x), 'float32')
        measurements = np.zeros((self.n_y, self.n_x, self.phase_steps, *self.feedback.data_shape))

        for y in range(self.n_y):
            for x in range(self.n_x):
                for p in range(self.phase_steps):
                    phase_pattern[y, x] = p * 2 * np.pi / self.phase_steps
                    self.slm.set_phases(phase_pattern)
                    self.feedback.trigger(out=measurements[y, x, p, ...])
                phase_pattern[y, x] = 0

        self.feedback.wait()
        starting_intensity = np.mean(measurements[:, :, 0])  # flat wavefront intensity
        return analyze_phase_stepping(measurements, axis=2, A=np.sqrt(starting_intensity))

    @property
    def n_x(self) -> int:
        """
        Returns:
            int: The number of SLM elements in the x direction.
        """
        return self._n_x

    @n_x.setter
    def n_x(self, value):
        self._n_x = value

    @property
    def n_y(self) -> int:
        """
        Returns:
            int: The number of SLM elements in the y direction.
        """
        return self._n_y

    @n_y.setter
    def n_y(self, value):
        self._n_y = value

    @property
    def phase_steps(self) -> int:
        """
        Returns:
            int: The number of phase steps used for each segment.
        """
        return self._phase_steps

    @phase_steps.setter
    def phase_steps(self, value):
        self._phase_steps = value



from . import microscope
from . import mockdevices

from .microscope import Microscope
from .mockdevices import MockXYStage, MockSource, MockCamera, MockSLM, ADCProcessor
from . import simulation
from .simulation import SimulatedWFS



from astropy import units as u
from matplotlib import pyplot as plt

from .core import Detector
from .utilities import get_pixel_size


def grab_and_show(cam: Detector, magnification=1.0):
    image = cam.read()
    dimensions = cam.extent.to_value(u.um) / magnification
    plt.imshow(image, extent=(0.0, dimensions[0], 0.0, dimensions[1]), cmap='gray')
    plt.xlabel('μm')
    plt.ylabel('μm')


def imshow(data):
    pixel_size = get_pixel_size(data)
    if pixel_size is not None:
        extent = pixel_size * data.shape
        plt.xlabel(extent.unit)
        plt.ylabel(extent.unit)
        extent = extent.value
    else:
        extent = data.shape
    plt.imshow(data, extent=(0.0, extent[0], 0.0, extent[1]), cmap='gray')
    plt.colorbar()
    plt.show()
    plt.pause(0.1)



from . import utilities
from . import patterns
from .utilities import ExtentType, CoordinateType, unitless, get_pixel_size, set_pixel_size, Transform, project, place
from .patterns import coordinate_range



import numpy as np
from typing import Optional
from ..utilities.patterns import gaussian
from .mockdevices import MockSLM, Processor


class SimulatedWFS(Processor):
    """A simple simulation of a wavefront shaping experiment.

    Simulates a configuration where both the SLM and the aberrations are placed in the back pupil
    of a microscope objective, and a point detector is placed in the the center of the focal plane.

    The simulation computes (Σ A·exp(i·(aberrations-slm)))², which is the 0,0 component of the Fourier transform
    of the field in the pupil plane (including aberrations and the correction by the SLM).

    For a more advanced (but slower) simulation, use `Microscope`
    """

    def __init__(self, aberrations: np.ndarray, beam_profile_waist=None):
        """
        Initializes the optical system with specified aberrations and optionally a Gaussian beam profile.

        This constructor sets up an optical system by specifying the aberrations across the SLM (Spatial Light
        Modulator) and, optionally, by defining a Gaussian beam profile. The aberrations array defines phase shifts at
        each point of the SLM. If a beam profile waist is provided, the input electric field is shaped into a
        Gaussian beam.

        Args:
            aberrations (np.ndarray): An array containing the aberrations in radians.

            beam_profile_waist (float, optional): The waist size of the Gaussian beam profile. If provided, the electric
            field at the SLM is shaped into a Gaussian beam with this waist size.

        The constructor creates a MockSLM instance based on the shape of the aberrations, calculates the electric
        field at the SLM considering the aberrations and optionally the Gaussian beam profile, and initializes the
        system with these parameters.
        """
        self.slm = MockSLM(aberrations.shape[0:2])
        self.E_input_slm = np.exp(1.0j * aberrations)  # electric field incident at the SLM
        if beam_profile_waist is not None:
            self.E_input_slm *= gaussian(aberrations.shape, waist=beam_profile_waist)

        # normalize the field
        self.E_input_slm *= 1 / np.linalg.norm(self.E_input_slm.ravel())
        super().__init__(self.slm.pixels())

    def _fetch(self, out: Optional[np.ndarray], slm_phases):  # noqa
        """
        Computes the intensity in the focus by applying phase corrections to the input electric field.

        This method adjusts the phase of the input electric field using the SLM (Spatial Light Modulator) phases,
        computes the tensor product with the electric field at the SLM, and then calculates the intensity of the
        resulting field.

        Args:
            out (Optional[np.ndarray]): An optional numpy array to store the calculated intensity. If provided, the
            intensity is stored in this array.

            slm_phases (np.ndarray): The phase corrections to apply, typically provided as a numpy array representing
            the phases set on the SLM.

        Returns:
            np.ndarray: A numpy array containing the calculated intensity in the focus.

        """
        correction = np.exp(1.0j * slm_phases)
        field = np.tensordot(correction, self.E_input_slm, 2)
        intensity = np.abs(field) ** 2
        if out is not None:
            out[...] = intensity
        return intensity

    @property
    def data_shape(self):
        return self.E_input_slm.shape[2:]



import numpy as np
from typing import Union, Sequence, Optional
from astropy.units import Quantity
from .utilities import ExtentType, CoordinateType, unitless

# shape of a numpy array, or a single integer that is broadcasted to a square shape
ShapeType = Union[int, Sequence[int]]

# a scalar quantity with optional unit attached
ScalarType = Union[float, np.ndarray, Quantity]

"""
Library of functions to create commonly used patterns

Each of the functions takes a `shape` input, which may be a scalar integer or a 2-element Sequence of integers
indicating the size (shape) in pixels of the returned field. If `shape` is a scalar, the same value is
used for both axes.

For the coordinates, the OpenGL convention is used, where the coordinates indicate the centers of the pixels.
By default, the returned pattern is assumed to cover a -1,1 x -1,1 square, 
which corresponds to a default `extent` parameter of (2.0, 2.0).
In this case, the coordinates range from -1+dx/2 to 1-dx/2, where dx=2.0/shape is the pixel size.

Optionally, a different `extent` can be specified to scale the coordinates.
This is especially useful when working with anisotropic pixels or rectangular patterns.
For example, a square pattern with anisotropic may be described by shape=(80,100) and extent(2,2)
whereas shape=(80,100) and extent(8,10) describes square pixels that form a rectangle.

In a pupil-conjugate configuration, the transformation matrix of the SLM should be set such that
SLM coordinates correspond to normalized pupil coordinates.
In this case, a disk of extent=(NA, NA) exactly covers the back pupil of the microscope objective.

The extent may have a unit of measure. In this case, other parameters (such as `radius`) may need to have
an according unit of measure.

The (0,0) coordinate is always located in the center of the pattern, which may be on a grid point (for odd shape)
or between grid points (for even shape).

The returned array has a pixel_size property attached.
"""


def coordinate_range(shape: ShapeType, extent: ExtentType, offset: Optional[CoordinateType] = None) -> (Quantity,
                                                                                                        Quantity):
    """
    Returns coordinate vectors for the two coordinates (y and x).

    Given a range from `-extent/2` to `+extent/2`, uniformly divided into `shape` pixels,
    the returned coordinates correspond to the centers of these coordinates.

    Arguments:
        shape (ShapeType): size of the full grid (y, x) in pixels
        extent (ExtentType): extent of the coordinate range
        offset (Optional[CoordinateType]): offset to be added to the coordinates (optional)

    Returns:
        Tuple[Quantity, Quantity]: coordinate vectors for the two coordinates (y and x)
    """
    if np.size(shape) == 1:
        shape = np.array(shape).repeat(2)

    extent = Quantity(extent)
    if extent.size == 1:
        extent = extent.repeat(2)

    if offset is None:
        offset = 0.0 * extent  # by default, let center be (0,0) (in whatever unit)

    def c_range(res, ex, cx):
        dx = ex / res
        return np.arange(res) * dx + (0.5 * dx - 0.5 * ex + cx)

    return (c_range(shape[0], extent[0], offset[0]).reshape((-1, 1)),
            c_range(shape[1], extent[1], offset[1]).reshape((1, -1)))


def r2_range(shape: ShapeType, extent: ExtentType):
    """Convenience function to return square distance to the origin
    Equivalent to computing cx^2 + cy^2
    """
    c0, c1 = coordinate_range(shape, extent)
    return c0 ** 2 + c1 ** 2


def tilt(shape: ShapeType, g: ExtentType, extent: ExtentType = (2.0, 2.0), phase_offset: float = 0.0):
    """Constructs a linear gradient pattern φ=2 g·r

    Args:
        shape: see module documentation
        g(tuple of two floats): gradient vector.
          This has the unit: 1 / extent.unit.
          For the default extent of (2.0, 2.0), a value of g=(1,0)
          corresponds to having a ramp from -2 to +2 over the height of the pattern
          When this pattern is used as a phase in a pupil-conjugate configuration,
          this corresponds to a displacement of -2/π times the Abbe diffraction limit
          (e.g. a positive x-gradient g causes the focal point to move in the _negative_ x-direction)
        extent: see module documentation
        phase_offset: optional additional phase offset to be added to the pattern
    """
    c0, c1 = coordinate_range(shape, extent * (Quantity(g) * 2.0))
    return unitless(c0 + (c1 + phase_offset))


def lens(shape: ShapeType, f: ScalarType, wavelength: ScalarType, extent: ExtentType):
    """Constructs a square texture that represents a wavefront defocus: (f-sqrt(f²+r²)) · 2π/λ

    `extent`, `wavelength` and `f` should have compatible units (typically astropy length units).

    Args:
        shape(ShapeType): see module documentation
        f(ScalarType): focal length
        wavelength(ScalarType): wavelength
        extent(ExtentType): physical extent of the SLM, same units as `f` and `wavelength`
    """
    r_sqr = r2_range(shape, extent)
    return unitless((f - np.sqrt(f ** 2 + r_sqr)) * (2 * np.pi / wavelength))


def propagation(shape: ShapeType, distance: ScalarType, numerical_aperture: ScalarType,
                refractive_index: ScalarType, wavelength: ScalarType, extent: ExtentType = (2.0, 2.0)):
    """Computes a wavefront that corresponds to digitally propagating the field in the object plane.

    k_z = sqrt(n² k_0²-k_x²-k_y²)
    φ = k_z · distance

    Args:
          shape: see module documentation
          distance (ScalarType): physical distance to propagate axially.
          numerical_aperture (Scalar):
          refractive_index (Scalar):
          wavelength (Scalar):
            the numerical aperture, refractive index and wavelength are used to
            to convert the `extent` from pupil coordinates to k-space (unit radians/meter),
          extent: extent of the returned image, in pupil coordinates
            (a disk of radius 1.0 corresponds to the full NA)
    """
    # convert pupil coordinates to absolute k_x, k_y coordinates
    k_0 = 2.0 * np.pi / wavelength
    extent_k = Quantity(extent) * numerical_aperture * k_0
    k_z = np.sqrt(np.maximum((refractive_index * k_0) ** 2 - r2_range(shape, extent_k), 0.0))
    return unitless(k_z * distance)


def disk(shape: ShapeType, radius: ScalarType = 1.0, extent: ExtentType = (2.0, 2.0)):
    """Constructs an image of a centered (ellipsoid) disk.

    (x / rx)^2 + (y / ry)^2 <= 1.0

    Args:
          shape: see module documentation
          radius (ScalarType): radius of the disk, should have the same unit as `extent`.
          extent: see module documentation
    """
    return 1.0 * (r2_range(shape, extent) < radius ** 2)


def gaussian(shape: ShapeType, waist: ScalarType,
             truncation_radius: ScalarType = None, extent: ExtentType = (2.0, 2.0)):
    """Constructs an image of a centered Gaussian

    `waist`, `extent` and the optional `truncation_radius` should all have the same unit.
    Arguments:
        shape: see module documentation
        waist (ScalarType):
            location of the beam waist (1/e value)
            relative to half of the size of the pattern (i.e. relative to the `radius` of the square)
        truncation_radius (ScalarType):
            when not None, specifies the radius of a disk that is used to truncate the Gaussian.
            All values outside the disk are set to 0.
        extent: see module documentation
    """
    r_sqr = r2_range(shape, extent)
    w2inv = -1.0 / waist ** 2
    gauss = np.exp(unitless(r_sqr * w2inv))
    if truncation_radius is not None:
        gauss = gauss * disk(shape, truncation_radius, extent=extent)
    return unitless(gauss)



import logging
import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Set, final, Tuple, Optional, Union
from weakref import WeakSet

import astropy.units as u
import numpy as np
from numpy.typing import ArrayLike
from astropy.units import Quantity

from .utilities import set_pixel_size


# Aliases for commonly used type hints

# An extent is a sequence of two floats with an optional unit attached,
# or a single float with optional unit attached, which is broadcasted to a sequence of two floats.


class Device(ABC):
    """Base class for detectors and actuators

    Multi-threading:
        Devices hold a thread lock, prevents concurrent modification of attributes
        in a multithreaded environment.
        This lock is automatically acquired when setting a public attribute or property.
        See `__setattr__` for details.

    Detector locking:
        For detectors, an additional synchonization mechanism is implemented.
        It prevents modification of the detector settings while a measurement is in progress.
        See Detector for details.

    Synchronization:
        Device implements the synchronization between detectors and actuators.
        The idea is that a measurement can only be made when all actuators are stable,
        and that an actuator can only be moved when all detectors are ready.

        The synchronization mechanism is implemented using a global state variable `_moving`.
        A detector can request a state switch to the `measuring` state (`_moving=False`),
        and an actuator can request a state switch to the `moving` state (`_moving=True`)
        by calling `_start`.

        Before making the state switch, `_start` waits for all devices of the _other_ type
        (actuators or detectors) to become ready by calling
        `wait(up_to=latency, await_data=False)` on those devices.
        Here, `latency` is the minimum latency of all devices of the _same_ type as the one
        requesting the state switch.
        If this minimum latency is positive, it means that the devices will not start their
        measurement or movement immediately, so we can make the state switch slightly before
        the devices of the other type are all ready.
        For example, a spatial light modulator has a relatively long latency, meaning that
        we can send the next frame even before a camera has finished reading the previous frame.

        For detectors, `_start` is called automatically by `trigger()`, so there is never a need to call it.
        Implementations of an actuator should call `_start` explicitly before starting to move the actuator.

    Attributes:
        multi_threading (bool): Option to globally disable multi-threading.
            This is particularly useful for debugging.
    Usage::
        >>> f1 = np.zeros((N, P, *cam1.data_shape))
        >>> f2 = np.zeros((N, P, *cam2.data_shape))
        >>> for n in range(N):
        >>>     for p in range(P)
        >>>         phase = 2 * np.pi * p / P
        >>>
        >>>         # wait for all measurements to complete (up to the latency of the slm), and trigger the slm.
        >>>         slm.set_phases(phase)
        >>>
        >>>         # wait for the image on the slm to stabilize, then trigger the measurement.
        >>>         cam1.trigger(out = f1[n, p, ...])
        >>>
        >>>         # directly trigger cam2, since we already are in the 'measuring' state.
        >>>         cam2.trigger(out = f2[n, p, ...])
        >>>
        >>> cam1.wait() # wait until camera 1 is done grabbing frames
        >>> cam2.wait() # wait until camera 2 is done grabbing frames
        >>> fields = (f2 - f1) * np.exp(-j * phase)

    """
    __slots__ = ('_end_time_ns', '_timeout_margin', '_lock', '_locking_thread', '_error', '_base_initialized',
                 '__weakref__')
    _workers = ThreadPoolExecutor(thread_name_prefix='Device._workers')
    _moving = False
    _state_lock = threading.Lock()
    _devices: "Set[Device]" = WeakSet()
    multi_threading: bool = True

    def __init__(self):
        """Constructs a new Device object"""
        self._end_time_ns = 0
        self._timeout_margin = 5 * u.s
        self._lock = threading.Lock()
        self._locking_thread = None
        self._error = None
        self._base_initialized = True
        Device._devices.add(self)

    def __del__(self):
        """
        Destructor for Device objects.

        The destructor calls `wait()` to ensure that the device is not busy when it is destroyed.
        If `__init__` did not complete, the destructor does nothing.

        Note:
            When closing Python, all modules are unloaded in an undefined order.
            This can cause problems if `wait` calls a function form a module that is already unloaded, such as `numpy`.
            In this case, a 'NoneType is not callable' error occurs.
        """
        if hasattr(self, '_base_initialized'):
            self.wait()

    def __setattr__(self, key, value):
        """Prevents modification of public attributes and properties while the device is locked.

        For detectors, this prevents modification of the detector settings while a measurement is in progress.
        For all devices, it prevents concurrent modification in a multi-threading context.

        Private attributes can be set without locking.
        Note that this is not thread-safe and should be done with care!!
        """
        if hasattr(self, '_base_initialized') and not key.startswith('_'):
            new_lock = self._lock_acquire()
            try:
                super().__setattr__(key, value)
            finally:
                if new_lock:
                    self._lock_release()
        else:
            super().__setattr__(key, value)

    @property
    @abstractmethod
    def _is_actuator(self):
        """True for actuators, False for detectors"""
        ...

    def _start(self):
        """Switches the state to 'moving' (for actuators) or 'measuring' (for detectors).

        This function changes the global state to 'moving' or 'measuring' if needed,
        and it may block until this state switch is completed.

        After switching, stores the time at which the operation will have ended in the `_end_time_ns`
        field (i.e., `time.time_ns() + self.latency + self.duration`).
        """

        # acquire a global lock, to prevent multiple threads to switch moving/measuring state simultaneously
        with Device._state_lock:
            # check if transition from moving/measuring or vice versa is needed
            if Device._moving != self._is_actuator():
                if Device._moving:
                    logging.debug("switch to MEASURING requested by %s.", self)
                else:
                    logging.debug("switch to MOVING requested by %s.", self)

                same_type = [device for device in Device._devices if device._is_actuator == self._is_actuator]
                other_type = [device for device in Device._devices if device._is_actuator != self._is_actuator]

                # compute the minimum latency of same_type
                # for instance, when switching to 'measuring', this number tells us how long it takes before any of the
                # detectors actually starts a measurement.
                # If this is a positive number, we can make the switch to 'measuring' slightly _before_
                # all actuators have stabilized.
                latency = min((device.latency for device in same_type), default=0.0 * u.ns)

                # wait until all devices of the other type have (almost) finished
                for device in other_type:
                    device.wait(up_to=latency, await_data=False)

                # changes the state from moving to measuring and vice versa
                Device._moving = not Device._moving
                if Device._moving:
                    logging.debug("state is now MOVING.")
                else:
                    logging.debug("state is now MEASURING.")

            # also store the time we expect the operation to finish
            # note: it may finish slightly earlier since (latency + duration) is a maximum value
            self._end_time_ns = time.time_ns() + self.latency.to_value(u.ns) + self.duration.to_value(u.ns)

    @property
    def latency(self) -> Quantity[u.ms]:
        """latency (Quantity[u.ms]): minimum amount of time between sending a command or trigger to the device and the
        moment the device starts responding.

        The default value is 0.0 ms.

        Note:
            The latency is used to compute when it is safe to switch the global state from 'moving'
            to 'measuring' or vice versa.
            Devices that report a non-zero latency promise not to do anything before the latency has passed.
            This allows making the state switch even before the devices of the other type have all finished.

            This construct is used for spatial light modulators, which typically have a long latency (1-2 frames).
            Due to this latency, we may send an `update` command to the SLM even before the camera has finished reading
            the previous frame.

        Note:
            A device is allowed to report a different latency every time `latency` is called.
            For example, for a spatial light modulator that only refreshes
            at a fixed rate, we can add the remaining time until the next refresh to the latency.
        """
        return 0.0 * u.ms

    @property
    @abstractmethod
    def duration(self) -> Quantity[u.ms]:
        """duration (Quantity[u.ms]): maximum amount of time it takes to perform the measurement or for the
        actuator to stabilize.
        This value *does not* include the latency.

        For a detector, this is the maximum amount of time that elapses between returning from `trigger()` and the
        end of the measurement.
        For an actuator, this is the maximum amount of time that elapses from returning from a command like `update(
        )` and the stabilization of the device.

        If the duration of an operation is not known in advance,
        (e.g., when waiting for a hardware trigger), this function should return `np.inf * u.ms`.

        Note: A device may update the duration dynamically.
        For example, a stage may compute the required time to
        move to the target position and update the duration accordingly.

        Note: If `latency` is a (lower) estimate, the duration should be high enough to guarantee that `latency +
        duration` is at least as large as the time between starting the operation and finishing it.
        """
        ...

    def wait(self, up_to: Quantity[u.ms] = 0 * u.ns, await_data=True) -> None:
        """Waits until the device is (almost) in the `ready` state, i.e., has finished measuring or moving.

        This function is called by `_start` automatically to ensure proper synchronization between detectors and
        actuators, and it is called by `__del__` to ensure the device is not active when it is destroyed.
        The only time to call `wait` explicitly is when using pipelined measurements, see `Detector.trigger()`.

        For devices that report a duration (`duration ≠ ∞`), this function waits until
        `current_time - up_to >= self._end_time_ns`,
        where `_end_time_ns` was set by the last call to `_start`.

        For devices that report no duration `duration = ∞`, this function repeatedly calls `busy` until `busy`
        returns `False`.
        In this case, `up_to` is ignored.

        Args:
            up_to(Quantity[u.ms]):
                when non-zero, specifies that this function may return 'up_to' milliseconds
                *before* the device is finished.
            await_data(bool):
                If True, after waiting until the device is no longer busy, briefly locks the device.
                For detectors, this has the effect of waiting until all acquisitions and processing of
                previously triggered frames are completed.
                If an `out` parameter was specified in `trigger()`, this guarantees that the data is stored
                in the `out` array.
                For actuators, this flag has no effect other than briefly locking the device.

        Raises:
            Any other exception raised by the device in another thread (e.g., during `_fetch`).
            TimeoutError: if the device has `duration = ∞`, and `busy` does not return `True` within
                `self.timeout`
            RuntimeError: if `wait` is called from inside a setter or from inside `_fetch`.
                This would cause a deadlock.

        """
        if self._error is not None:
            e = self._error
            self._error = None
            raise e

        # If duration = ∞, poll busy until it returns False or a timeout occurs.
        # Note: avoid np.isinf because numpy may have been unloaded during Python shutdown
        if self._end_time_ns > 1.0e+38:
            start = time.time_ns()
            timeout = self.timeout.to_value(u.ns)
            while self.busy():
                time.sleep(0.001)
                if time.time_ns() - start > timeout:
                    raise TimeoutError("Timeout in %s (tid %i)", self, threading.get_ident())
        else:
            time_to_wait = self._end_time_ns - up_to.to_value(u.ns) - time.time_ns()
            if time_to_wait > 0:
                time.sleep(time_to_wait / 1.0E9)

        if await_data:
            # locks the device, for detectors this waits until all pending measurements are processed.
            if not self._lock_acquire():
                raise RuntimeError(
                    "Cannot call `wait` from inside a setter or from inside _fetch as it will cause a deadlock.")
            self._lock_release()

    def _lock_acquire(self) -> bool:
        """Acquires a non-persistent lock using the timeout of the device"""
        tid = threading.get_ident()
        if self._locking_thread == tid:
            return False  # already locked, this could happen if we call a setter from _fetch or from another setter
        if not self._lock.acquire(timeout=self.timeout.to_value(u.s)):
            raise TimeoutError("Timeout in %s (tid %i)", self, tid)
        self._locking_thread = tid
        return True

    def _lock_release(self):
        """Releases a non-persistent lock"""
        self._locking_thread = None
        self._lock.release()

    def busy(self) -> bool:
        """Returns true if the device is measuring or moving (see `wait()`).

        Note: if a device does not define a finite `duration`, it must override this function to poll for
        finalization."""
        return time.time_ns() < self._end_time_ns

    @property
    def timeout(self) -> Quantity[u.ms]:
        """Time after which a timeout error is raised when waiting for the device.

        The timeout is automatically adjusted if the `duration` changes.
        The default value is `duration + 5 s`."""
        duration = self.duration
        if not np.isinf(duration):
            return self._timeout_margin + duration
        else:
            return self._timeout_margin

    @timeout.setter
    def timeout(self, value):
        duration = self.duration
        if not np.isinf(duration):
            self._timeout_margin = value - duration
        else:
            self._timeout_margin = value.to(u.ms)


class Actuator(Device, ABC):
    """Base class for all actuators
    """
    __slots__ = ()

    @final
    def _is_actuator(self):
        return True


class Detector(Device, ABC):
    """Base class for all detectors, cameras and other data sources with possible dynamic behavior.
    """
    __slots__ = ('_measurements_pending', '_pending_count_lock')

    def __init__(self):
        super().__init__()
        self._measurements_pending = 0
        self._pending_count_lock = threading.Lock()
        self._error = None

    @final
    def _is_actuator(self):
        return False

    def _lock_persistent(self):
        """Places a persistent lock on the detector.

        A persistent lock is a lock that is only released after all measurements have finished.

        If the detector was not locked, it is locked, and the `measurements_pending` counter is incremented.
        Only when the measurements_pending counter drops to zero, the lock is released (see `_unlock_persistent()`).

        If the detector was already locked, either wait for the lock to be released
        (if the existing lock was not persistent),
        or just increment the `measurements_pending` counter
        (if the existing lock was persistent).
        """
        logging.debug("entering persistent lock %s (tid %i).", self, threading.get_ident())
        with self._pending_count_lock:
            if self._measurements_pending == 0:
                if not self._lock_acquire():  # we don't have a persistent lock yet, acquire the object lock
                    raise RuntimeError("Cannot call `trigger` from a setter or from _fetch")
                self._locking_thread = None
                logging.debug("first persistent lock set %s (tid %i).", self, threading.get_ident())
            self._measurements_pending += 1

    def _unlock_persistent(self):
        """Decrement the `measurements_pending` counter, and release the persistent lock if it reached 0."""
        logging.debug("leaving persistent lock %s (tid %i).", self, threading.get_ident())
        with self._pending_count_lock:
            self._measurements_pending -= 1
            if self._measurements_pending == 0:  # we released the last persistent lock
                self._lock_release()
                logging.debug("last persistent lock cleared %s (tid %i).", self, threading.get_ident())

    def trigger(self, *args, out=None, immediate=False, **kwargs) -> Future:
        """Triggers the detector to start acquisition of the data.

        This function returns a `concurrent.futures.Future`.
        Call `.result()` on the returned object to wait for the data.

        All parameters are passed to the _fetch function of the detector.
        If any of these parameters is a Future, it is awaited before calling _fetch.
        This way, data from multiple sources can be combined (see Processor).

        Child classes may override trigger() to call `super().trigger()` with additional parameters.

        Note:
            To implement hardware triggering, do not override this function.
            Instead, override `_do_trigger()` instead to ensure proper synchronization and locking.
        """

        self._lock_persistent()

        # if the detector is not locked yet,
        # lock it now (this is a persistent lock),
        # which is released in _do_fetch if the number of pending measurements reaches zero
        # does nothing if the lock is already locked.
        try:
            self._start()
            self._do_trigger()
        except:  # noqa  - ok, we are not really catching the exception, just making sure the lock gets released
            self._unlock_persistent()
            raise

        logging.debug("triggering %s (tid: %i).", self, threading.get_ident())
        if immediate or not Device.multi_threading:
            result = Future()
            result.set_result(self.__do_fetch(out, *args, **kwargs))  # noqa
            return result
        else:
            return Device._workers.submit(self.__do_fetch, out, *args, **kwargs)

    def __do_fetch(self, out_, *args_, **kwargs_):
        """Helper function that awaits all futures in the keyword argument list, and then calls _fetch"""
        try:
            if len(args_) > 0 or len(kwargs_) > 0:
                logging.debug("awaiting inputs for %s (tid: %i).", self, threading.get_ident())
            awaited_args = [(arg.result() if isinstance(arg, Future) else arg) for arg in args_]
            awaited_kwargs = {key: (arg.result() if isinstance(arg, Future) else arg) for (key, arg) in
                              kwargs_.items()}
            logging.debug("fetching data of %s ((tid: %i)).", self, threading.get_ident())
            data = self._fetch(out_, *awaited_args, **awaited_kwargs)
            data = set_pixel_size(data, self.pixel_size)
            assert data.shape == self.data_shape
            return data
        except Exception as e:
            # if we are storing the result in an `out` array,
            # the user may never call result() on the returned future object,
            # and the error may be lost.
            # Therefore, store it so that it can be raised on the next call to wait()
            if out_ is not None:
                self._error = e
            raise e  # raise the error again, it will be stored in the 'Future' object that was returned by trigger()
        finally:
            self._unlock_persistent()

    def _do_trigger(self) -> None:
        """Override this function to perform the actual hardware trigger."""
        pass

    @abstractmethod
    def _fetch(self, out: Optional[np.ndarray], *args, **kwargs) -> np.ndarray:
        """Read the data from the detector

        Args:
            out(ndarray) numpy array or view of an array that will receive the data
                when present, the data will be stored in `out`, and `out` is returned.
                Otherwise, a new array is returned.
                Can also be used with views.
                For example, `trigger(out = data[i, ...])`
            The args and kwargs are passed from the call to trigger()
        Note:
            After reading the data, the Detector attaches the pixel_size metadata.
        Note:
            Child classes must implement this function, and store the data in `out[...]` if `out` is not None.
        """
        ...

    @final
    def read(self, *args, immediate=True, **kwargs) -> np.ndarray:
        """Triggers the detector and waits for the data to arrive.

        Shortcut for trigger().result().
        """
        return self.trigger(*args, immediate=immediate, **kwargs).result()

    @property
    @abstractmethod
    def data_shape(self) -> Tuple[int, ...]:
        """The shape of the data array that `read()` will return.

        For some detectors this property may be mutable, for example, for a camera it
        represents the height and width of the ROI, which can be changed.
        """
        ...

    @property
    def pixel_size(self) -> Optional[Quantity]:
        """Physical dimension of one element in the returned data array.

        For cameras, this is the pixel size (in astropy length units).
        For detectors returning a time trace, this value is specified in astropy time units.

        The pixel_size is a 1-D array with an element for each dimension of the returned data.

        By default, the pixel size cannot be set.
        However, in some cases (such as when the `pixel_size` is actually a sampling interval),
        it makes sense for the child class to implement a setter.
        """
        return None

    @final
    def coordinates(self, dim: int) -> Quantity:
        """Returns an array with the coordinate values along the d-th axis.

        The coordinates represent the center of the pixels (or sample intervals
        in case the detector returns a time signal).
        Therefore, the coordinates range from `0.5 * pixel_size` to `(data_shape[d]-0.5) * pixel_size`.
        Coordinates have the same astropy base unit as pixel_size.
        The coordinates are returned as an array with the same number of dimensions as the returned data,
        with the d-th dimension holding the coordinates.
        This faclilitates meshgrid-like computations, e.g.
        `cam.coordinates(0) + cam.coordinates(1)` gives a 2-dimensional array of coordinates.

        Args:
            dim: Dimension for which to return the coordinates.
        """
        unit = u.dimensionless_unscaled if self.pixel_size is None else self.pixel_size[dim]
        c = np.arange(0.5, 0.5 + self.data_shape[dim], 1.0) * unit
        shape = np.ones_like(self.data_shape)
        shape[dim] = self.data_shape[dim]
        return c.reshape(shape)

    @final
    @property
    def extent(self) -> Quantity:
        """Physical size of the data array

        If a `pixel_size` is set, this function returns `data_shape * pixel_size`
        as an `astropy.units.Quantity`.
        If no `pixel_size` is set, this function uses  the
        `dimensionless_unscaled` unit.
        """
        unit = u.dimensionless_unscaled if self.pixel_size is None else self.pixel_size
        return np.array(self.data_shape) * unit


class Processor(Detector, ABC):
    """Base class for all Processors.

    Processors can be used to build data processing graphs, where each Processor takes input from one or
    more input Detectors and processes that data (e.g., cropping an image, averaging over an ROI, etc.).
    A processor, itself, is a Detector to allow chaining multiple processors together to combine functionality.

    To implement a processor, implement `_fetch`, and optionally override `data_shape`, `pixel_size`, and `__init__`.
    The `latency` and `duration` properties are computed from the latency and duration of the inputs and cannot be set.
    By default, the `pixel_size` and `data_shape` are the same as the `pixel_size` and `data_shape` of the first input.
    To override this behavior, override the `pixel_size` and `data_shape` properties.
    """

    def __init__(self, *args):
        self._sources = args
        super().__init__()

    def trigger(self, *args, immediate=False, **kwargs):
        """Triggers all sources at the same time (regardless of latency), and schedules a call to `_fetch()`"""
        future_data = [(source.trigger(immediate=immediate) if source is not None else None) for source in
                       self._sources]
        return super().trigger(*future_data, *args, **kwargs)

    @final
    @property
    def latency(self) -> Quantity[u.ms]:
        """Returns the shortest latency for all detectors."""
        return min((source.latency for source in self._sources if source is not None), default=0.0 * u.ms)

    @final
    @property
    def duration(self) -> Quantity[u.ms]:
        """Returns the last end time minus the first start time for all detectors
        i.e., max (duration + latency) - min(latency).

        Note that `latency` is allowed to vary over time for devices that can only be triggered periodically,
        so this `duration` may also vary over time.
        """
        times = [(source.duration, source.latency) for source in self._sources if source is not None]
        if len(times) == 0:
            return 0.0 * u.ms
        return (max([duration + latency for (duration, latency) in times])
                - min([latency for (duration, latency) in times]))

    @property
    def data_shape(self):
        """This default implementation returns the data shape of the first source."""
        return self._sources[0].data_shape

    @property
    def pixel_size(self) -> Optional[Quantity]:
        """This default implementation returns the pixel size of the first source."""
        return self._sources[0].pixel_size


class PhaseSLM(ABC):
    """Base class for phase-only SLMs
    """
    __slots__ = ()

    @abstractmethod
    def update(self) -> None:
        """Sends the new phase pattern to be displayed on the SLM.

        Implementations should call _start() before triggering the SLM.

        Note:
            This function *does not* wait for the image to appear on the SLM.
            To wait for the image stabilization explicitly, use 'wait()'.
            However, this should rarely be needed since all Detectors
            already wait for the image to stabilize before starting a measurement.
        """

    @abstractmethod
    def set_phases(self, values: ArrayLike, update: bool = True) -> None:
        """Sets the phase pattern on the SLM.

        Args:
            values(ArrayLike): phase pattern, in radians.
                The pattern is automatically stretched to fill the full SLM.
            update: when True, calls `update` after setting the phase pattern.
                Set to `False` to suppress the call to `update`.
                This is useful in advanced scenarios where multiple parameters of the SLM need to be changed
                before updating the displayed image.
        """
        ...



import numpy as np
import weakref
from OpenGL.GL import *


class Texture:
    def __init__(self, context, texture_type=GL_TEXTURE_2D):
        self.context = weakref.ref(context)
        self.handle = glGenTextures(1)
        self.type = texture_type
        self.synchronized = False  # self.data is not yet synchronized with texture in GPU memory
        self._data_shape = None  # current size of the texture, to see if we need to make a new texture or
        # overwrite the exiting one

        # create a single pixel texture as default (also activates the OpenGL context and binds the texture
        self.set_data(0)

        # set wrapping and interpolation options
        glTexParameteri(self.type, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(self.type, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glTexParameteri(self.type, GL_TEXTURE_WRAP_R, GL_REPEAT)
        glTexParameteri(self.type, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(self.type, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

    def __del__(self):
        context = self.context()
        if context is not None and hasattr(self, 'handle'):
            context.activate()
            glDeleteTextures(1, [self.handle])

    def bind(self, idx):
        """ Bind texture to texture unit idx. Assumes that the OpenGL context is already active."""
        glActiveTexture(GL_TEXTURE0 + idx)
        glBindTexture(self.type, self.handle)

    def set_data(self, value):
        """ Set texture data.

        The texture data is directly copied to the GPU memory,
         so the original data array can be modified or deleted.
        """
        value = np.array(value, dtype=np.float32, order='C', copy=False)

        self.context().activate()
        glBindTexture(self.type, self.handle)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 4)  # alignment is at least four bytes since we use float32
        (internal_format, data_format, data_type) = (GL_R32F, GL_RED, GL_FLOAT)

        if self.type == GL_TEXTURE_1D:
            # check if data has the correct dimension, convert scalars to arrays of correct dimension
            if value.ndim == 0:
                value = value.reshape((1,))
            elif value.ndim != 1:
                raise ValueError("Data should be a 1-d array or a scalar")
            if value.shape != self._data_shape:
                # create a new texture
                glTexImage1D(GL_TEXTURE_1D, 0, internal_format, value.shape[0], 0, data_format, data_type, value)
                self._data_shape = value.shape
            else:
                # overwrite existing texture
                glTexSubImage1D(GL_TEXTURE_1D, 0, 0, value.shape[0], data_format, data_type, value)

        elif self.type == GL_TEXTURE_2D:
            if value.ndim == 0:
                value = value.reshape((1, 1))
            elif value.ndim != 2:
                raise ValueError("Data should be a 2-d array or a scalar")
            if value.shape != self._data_shape:
                glTexImage2D(GL_TEXTURE_2D, 0, internal_format, value.shape[1], value.shape[0], 0,
                             data_format, data_type, value)
                self._data_shape = value.shape
            else:
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, value.shape[1], value.shape[0], data_format,
                                data_type, value)
        else:
            raise ValueError("Texture type not supported")

    def get_data(self):
        self.context().activate()
        data = np.empty(self._data_shape, dtype='float32')
        glGetTextureImage(self.handle, 0, GL_RED, GL_FLOAT, data.size * 4, data)
        return data
